local VERSION = tonumber(("$Revision: 31607 $"):match("%d+"))

local SharedMedia = AceLibrary("SharedMedia-1.0")

PitBull = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceEvent-2.0", "AceConsole-2.0", "AceModuleCore-2.0", "FuBarPlugin-2.0")
local PitBull, self = PitBull, PitBull
PitBull.version = "r" .. VERSION
PitBull.revision = VERSION
PitBull.date = ("$Date: 2007-04-02 03:36:21 -0400 (Mon, 02 Apr 2007) $"):match("%d%d%d%d%-%d%d%-%d%d")

local DEBUG = false -- this enables profiling for this addon.

local L = setmetatable({}, {__index = function(self, key) self[key] = key; return key end })
-- this is used as a placeholder for any localization I choose to do later.

local configMode = false
PitBull.configMode = configMode

local wrap
do
	local start = GetTime()
	if DEBUG then
		local tree = {}
		local treeMemories = {}
		local treeTimes = {}
		local memories = {}
		local times = {}
		local calls = {}
		function wrap(value, name)
			if type(value) == "function" then
				local oldFunction = value
				memories[name] = 0
				times[name] = 0
				calls[name] = 0
				local function func(pos, t, mem, ...)
					mem, t = collectgarbage('count') - mem, GetTime() - t
					if pos > 0 then
						treeMemories[pos] = treeMemories[pos] + mem
						treeTimes[pos] = treeTimes[pos] + t
					end
					local otherMem = table.remove(treeMemories)
					if mem - otherMem > 0 then
						memories[name] = memories[name] + mem - otherMem
					end
					times[name] = times[name] + t - table.remove(treeTimes)
					table.remove(tree)
					return ...
				end
				return function(...)
					local pos = #tree
					tree[#tree+1] = name
					treeMemories[#treeMemories+1] = 0
					treeTimes[#treeTimes+1] = 0
					calls[name] = calls[name] + 1
					local t, mem = GetTime(), collectgarbage('count')
					return func(pos, t, mem, oldFunction(...))
				end
			elseif type(value) == "table" then
				for k, v in pairs(value) do
					if type(k) == "string" and type(v) == "function" and v ~= wrap then
						value[k] = wrap(v, name .. ':' .. k)
					end
				end
				return value
			end
		end

		function DEBUG()
			return GetTime() - start, times, memories, calls
		end
	else
		function wrap(value)
			return value
		end
	end
end
PitBull.wrap = wrap

local table_remove = table.remove
local select = select

local Dewdrop = AceLibrary("Dewdrop-2.0")
local Tablet = AceLibrary("Tablet-2.0")

local new, del, newHash
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return { ... }
		end
	end
	
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	
	function del(t)
		for k in pairs(t) do
			t[k] = nil
		end
		list[t] = true
		return nil
	end
end

local newFrame, delFrame
do
	local frameCache = {}
	local frameCount = {}
	function newFrame(kind, parent, extra)
		local frame
		if frameCache[kind] then
			frame = table_remove(frameCache[kind])
			if #frameCache[kind] == 0 then
				frameCache[kind] = del(frameCache[kind])
			end
			frame:SetParent(parent)
			if kind == "Texture" or kind == "FontString" then
				frame:SetDrawLayer(extra)
			end
			frame:Show()
		else	
			frameCount[kind] = (frameCount[kind] or 0) + 1
			local name = "PitBull" .. kind .. frameCount[kind]
			if kind == "Texture" then
				frame = parent:CreateTexture(name, extra)
			elseif kind == "FontString" then
				frame = parent:CreateFontString(name, extra)
			else
				frame = CreateFrame(kind, name, parent)
			end
		end
		return frame
	end
	
	function delFrame(frame)
		if type(frame) ~= "table" then
			error(("Bad argument #1 to `delFrame'. %s expected, got %s"):format("table", type(frame)), 2)
		end
		local kind = frame:GetObjectType()
		if kind == "FontString" then
			PitBull:UnregisterWatchdogFontString(frame)
			frame:SetText("")
			frame:SetFontObject(nil)
			frame:SetNonSpaceWrap(true)
			frame:SetJustifyH("LEFT")
			frame:SetTextColor(1, 1, 1, 1)
		elseif kind == "Texture" then
			frame:SetTexture(nil)
			frame:SetVertexColor(1, 1, 1, 1)
			frame:SetBlendMode("BLEND")
			frame:SetDesaturated(false)
			frame:SetTexCoord(0, 1, 0, 1)
			frame:SetTexCoordModifiesRect(false)
		elseif kind == "StatusBar" then
			frame:SetStatusBarColor(1, 1, 1, 1)
			frame:SetStatusBarTexture(nil)
			frame:SetValue(1)
			frame:SetOrientation("HORIZONTAL")
		end
		frame:Hide()
		frame:SetParent(UIParent)
		frame:ClearAllPoints()
		frame:SetAlpha(1)
		local frameCache_kind = frameCache[kind]
		if not frameCache_kind then
			frameCache_kind = new()
			frameCache[kind] = frameCache_kind
		end
		frameCache_kind[#frameCache_kind+1] = frame
		return nil
	end
end

PitBull.new = new
PitBull.newHash = newHash
PitBull.del = del
PitBull.newFrame = newFrame
PitBull.delFrame = delFrame

local UnitIsUnit = UnitIsUnit
local UnitPowerType = UnitPowerType
local table_remove = table.remove

local frames = {}
local framesByGroup = {}
local nonWackyFrames = {}
local allFrames = {}
local allFramesByGroup = {}
local allFramesByCluster = {}
local groupMenus = {}
local wackyUnits = {n = 0}
local layouts, layoutNames = {}, {}
local borders, borderNames = {}, {}
local frameToModule = {}
local moduleOptionsMethods = {}

local ProfilePitBull

PitBull:RegisterDB("PitBullDB")
do
	local colors = {
		rage = { 226/255, 45/255, 75/255 },
		energy = { 1, 220/255, 25/255 },
		focus = { 1, 210/255, 0 },
		mana = { 48/255, 113/255, 191/255 },
		
		unknown = { 0.8, 0.8, 0.8 },
		
		hostile = { 226/255, 45/255, 75/255 },
		neutral = { 1, 1, 34/255 },
		friendly = { 0.2, 0.8, 0.15 },
		civilian = { 48/255, 113/255, 191/255 },
		
		dead = { 0.6, 0.6, 0.6 },
		disconnected = { 1, 0.5, 0 },
		inCombat = { 1, 0, 0 },
		resting = { 1, 1, 0 },
		tapped = { 0.5, 0.5, 0.5 },
		
		petHappy = { 0, 1, 0 },
		petNeutral = { 1, 1, 0 },
		petAngry = { 1, 0, 0 },
		
		frameBG = { 0, 0, 0, 0.5 },
		frameBorder = { 1, 0.7, 1, 1 },
	}
	
	for k, v in pairs(RAID_CLASS_COLORS) do
		colors[k] = { v.r, v.g, v.b }
	end
	
	PitBull:RegisterDefaults('profile', {
		settings = {
			texture = "Blizzard",
			locked = false,
			clamped = true,
			colors = colors,
			hidePartyInRaid = true,
		},
		groups = {
			['**'] = {
				scale = 1,
				width = 160,
				height = 50,
				layout = "ABF",
				border = "Hidden",
				hidden = false,
				direction = "down",
				spacing = 10,
				freeform = false,
			},
			targettarget = {
				height = 30,
			},
			targettargettarget = {
				height = 30,
				hidden = true,
			},
			pettarget = {
				hidden = true,
			},
			partytarget = {
				hidden = true,
			},
			raidtarget = {
				hidden = true,
			},
			raidpet = {
				hidden = true,
			},
			mouseover = {
				hidden = true,
			}
		},
		units = {
			['*'] = {
				x = 0,
				y = 0,
			}
		},
	})
end

local IsLegitimateUnit = {player=true,pet=true,target=true,focus=true,mouseover=true}
for i = 1, 4 do
	IsLegitimateUnit['party' .. i] = true
	IsLegitimateUnit['partypet' .. i] = true
end
for i = 1, 40 do
	IsLegitimateUnit['raid' .. i] = true
	IsLegitimateUnit['raidpet' .. i] = true
end
setmetatable(IsLegitimateUnit, {__index=function(self, unit)
	if not unit:find("target$") then
		self[unit] = false
		return false
	end
	local nonTarget = unit:sub(1, -7)
	local good = self[nonTarget]
	self[unit] = good
	return good
end})
PitBull.IsLegitimateUnit = IsLegitimateUnit

local IsWackyUnit = setmetatable({player=false, target=false, pet=false, mouseover=false, focus=false}, {__index=function(self, unit)
	local party = unit:match("^party(%d)$") or unit:match("^partypet(%d)$")
	if party then
		party = tonumber(party)
		if party and party >= 1 and party <= 4 then
			self[unit] = false
			return false
		end
		self[unit] = true
		return true
	end
	local raid = unit:match("^raid(%d%d?)$")
	if raid then
		raid = tonumber(raid)
		if raid and raid >= 1 and raid <= 40 then
			self[unit] = false
			return false
		end
		self[unit] = true
		return true
	end
	self[unit] = true
	return true
end})
PitBull.IsWackyUnit = IsWackyUnit

local GroupToLocale = {player = L["Player"], target = L["Target"], pet = L["Player's pet"], party = L["Party"], partypet = L["Party pets"], raid = L["Raid"], raidpet = L["Raid pets"], mouseover = L["Mouse-over"], focus = L["Focus"]}
setmetatable(GroupToLocale, {__index=function(self, group)
	if not group:find("target$") then
		self[group] = group
		return group
	end
	local nonTarget = group:sub(1, -7)
	local good
	if group:find("^player") or group:find("^pet") or group:find("^mouseover") or group:find("^target") or group:find("^focus") then
		good = L["%s's target"]:format(self[nonTarget])
	else
		good = L["%s targets"]:format(self[nonTarget])
	end
	self[group] = good
	return good
end})
PitBull.GroupToLocale = GroupToLocale

-- FuBar configuration
PitBull.hasIcon = true
PitBull.hideWithoutStandby = true
PitBull.defaultPosition = "MINIMAP"
PitBull.defaultMinimapPosition = 200
PitBull.hasNoText = true
PitBull.hasNoColor = true
PitBull.cannotDetachTooltip = true
PitBull.independentProfile = true

function PitBull:OnInitialize()
	self.colorConstants = self.db.profile.settings.colors
end

do
	local function colorSet(key, r, g, b, a)
		local t = self.db.profile.settings.colors[key]
		t[1] = r
		t[2] = g
		t[3] = b
		t[4] = a
		self:UpdateAll()
	end

	local function colorGet(key)
		return unpack(self.db.profile.settings.colors[key])
	end

	PitBull.options = {
		type = 'group',
		handler = PitBull,
		args = {
			new = {
				name = "New unit frame",
				desc = "Create a new unit frame",
				type = 'text',
				usage = "something like {focustarget || targettargettarget || etc.}",
				get = false,
				set = "CreateUnitFrame",
				validate = function(value)
					return IsLegitimateUnit[value]
				end,
				disabled = InCombatLockdown,
			},
			global = {
				name = "Global settings",
				desc = "Global settings",
				type = 'group',
				args = {
					texture = {
						name = "StatusBar texture",
						desc = "StatusBar texture",
						type = 'text',
						validate = SharedMedia:List('statusbar'),
						get = function()
							return self.db.profile.settings.texture
						end,
						set = function(value)
							self.db.profile.settings.texture = value
							self:UpdateAllStatusBarTextures()
						end,
					},
					locked = {
						name = "Locked",
						desc = "Make it so the frames cannot be moved",
						type = 'toggle',
						get = function()
							return self.db.profile.settings.locked
						end,
						set = function(value)
							self.db.profile.settings.locked = value
						end,
					},
					clamped = {
						name = "Clamp frames to screen",
						desc = "Make it so that frames cannot be dragged off-screen",
						type = 'toggle',
						get = "AreUnitFramesClampedToScreen",
						set = "ToggleUnitFramesClampedToScreen",
						handler = self,
						disabled = InCombatLockdown,
					},
					configMode = {
						name = "Configuration mode",
						desc = "Show all frames that can be shown, for easy configuration.",
						type = 'text',
						validate = {
							solo = "Solo",
							party = "Party",
							raid = "Raid",
							disabled = "Disabled"
						},
						get = function()
							return configMode or "disabled"
						end,
						set = "ChangeConfigMode",
						handler = self,
						disabled = InCombatLockdown,
					},
					hidePartyInRaid = {
						name = "Hide party frames in raid",
						desc = "Hide party frames in raid",
						type = 'toggle',
						get = function()
							return self.db.profile.settings.hidePartyInRaid
						end,
						set = "ChangeHidePartyInRaid",
						handler = self,
						disabled = InCombatLockdown,
					},
					colors = {
						name = "Colors",
						desc = "Colors",
						type = 'group',
						args = {
							power = {
								name = "Power",
								desc = "Power",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									rage = {
										name = "Rage",
										desc = "Rage",
										type = 'color'
									},
									mana = {
										name = "Mana",
										desc = "Mana",
										type = 'color'
									},
									focus = {
										name = "Focus",
										desc = "Focus",
										type = 'color'
									},
									energy = {
										name = "Energy",
										desc = "Energy",
										type = 'color'
									},
								}
							},
							class = {
								name = "Classes",
								desc = "Classes",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									WARLOCK = {
										name = "Warlock",
										desc = "Warlock",
										type = 'color'
									},
									PRIEST = {
										name = "Priest",
										desc = "Priest",
										type = 'color'
									},
									WARRIOR = {
										name = "Warrior",
										desc = "Warrior",
										type = 'color'
									},
									PALADIN = {
										name = "Paladin",
										desc = "Paladin",
										type = 'color'
									},
									SHAMAN = {
										name = "Shaman",
										desc = "Shaman",
										type = 'color'
									},
									MAGE = {
										name = "Mage",
										desc = "Mage",
										type = 'color'
									},
									DRUID = {
										name = "Druid",
										desc = "Druid",
										type = 'color'
									},
									ROGUE = {
										name = "Rogue",
										desc = "Rogue",
										type = 'color'
									},
									HUNTER = {
										name = "Hunter",
										desc = "Hunter",
										type = 'color'
									},
								}
							},
							reaction = {
								name = "Reaction",
								desc = "Reaction",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									hostile = {
										name = "Hostile",
										desc = "Hostile",
										type = 'color',
									},
									neutral = {
										name = "Neutral",
										desc = "Neutral",
										type = 'color',
									},
									friendly = {
										name = "Friendly",
										desc = "Friendly",
										type = 'color',
									},
									civilian = {
										name = "Civilian",
										desc = "Civilian",
										type = 'color',
									},
								}
							},
							petHappiness = {
								name = "Pet happiness",
								desc = "Pet happiness",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									petHappy = {
										name = "Happy",
										desc = "Happy",
										type = 'color',
									},
									petNeutral = {
										name = "Neutral",
										desc = "Neutral",
										type = 'color',
									},
									petAngry = {
										name = "Angry",
										desc = "Angry",
										type = 'color',
									},
								}
							},
							other = {
								name = "Other",
								desc = "Other",
								type = 'group',
								pass = true,
								get = colorGet,
								set = colorSet,
								args = {
									dead = {
										name = "Dead or Ghost",
										desc = "Dead or Ghost",
										type = 'color',
									},
									disconnected = {
										name = "Disconnected",
										desc = "Disconnected",
										type = 'color',
									},
									inCombat = {
										name = "In combat",
										desc = "In combat",
										type = 'color',
									},
									resting = {
										name = "Resting",
										desc = "Resting",
										type = 'color',
									},
									tapped = {
										name = "Tapped",
										desc = "Tapped",
										type = 'color',
									},
									unknown = {
										name = "Unknown",
										desc = "Unknown",
										type = 'color'
									},
								}
							},
							frame = {
								name = "Frame colors",
								desc = "Frame colors",
								type = 'group',
								args = {
									background = {
										name = "Background",
										desc = "Background",
										type = 'color',
										get = function()
											return unpack(self.db.profile.settings.colors.frameBG)
										end,
										set = function(r, g, b, a)
											local t = self.db.profile.settings.colors.frameBG
											t[1] = r
											t[2] = g
											t[3] = b
											t[4] = a
											self:UpdateAllBackdropsAndBorders()
										end,
										hasAlpha = true,
									},
									border = {
										name = "Border",
										desc = "Border",
										type = 'color',
										get = function()
											return unpack(self.db.profile.settings.colors.frameBorder)
										end,
										set = function(r, g, b, a)
											local t = self.db.profile.settings.colors.frameBorder
											t[1] = r
											t[2] = g
											t[3] = b
											t[4] = a
											self:UpdateAllBackdropsAndBorders()
										end,
										hasAlpha = true,
									},
								}
							}
						}
					},
				}
			},	
			debug = {
				order = -1,
				name = "Profile PitBull",
				desc = "Show the memory and times for each function of PitBull",
				type = 'execute',
				func = function() ProfilePitBull() end,
				hidden = function() return not ProfilePitBull end,
			},
		},
	}
end
PitBull.OnMenuRequest = PitBull.options

local modulesWithFunction = setmetatable({}, {__index = function(self, key)
	local t = {}
	for name, module in PitBull:IterateModules() do
		if type(module[key]) == "function" then
			t[#t+1] = module
		end
	end
	self[key] = t
	
	return t
end})
function PitBull:OnEnable(first)
	if not ClickCastFrames then
		ClickCastFrames = {}
	end
	
	Dewdrop:InjectAceOptionsTable(self, self.options)
	
	self:AddGroupToAceOptions("player")
	if not self.db.profile.groups.player.hidden then
		self:CreateUnitFrame("player")
	end
	self:AddGroupToAceOptions("pet")
	if not self.db.profile.groups.pet.hidden then
		self:CreateUnitFrame("pet")
	end
	self:AddGroupToAceOptions("pettarget")
	if not self.db.profile.groups.pettarget.hidden then
		self:CreateUnitFrame("pettarget")
	end
	self:AddGroupToAceOptions("target")
	if not self.db.profile.groups.target.hidden then
		self:CreateUnitFrame("target")
	end
	self:AddGroupToAceOptions("targettarget")
	if not self.db.profile.groups.targettarget.hidden then
		self:CreateUnitFrame("targettarget")
	end
	self:AddGroupToAceOptions("targettargettarget")
	if not self.db.profile.groups.targettargettarget.hidden then
		self:CreateUnitFrame("targettargettarget")
	end
	self:AddGroupToAceOptions("focus")
	if not self.db.profile.groups.focus.hidden then
		self:CreateUnitFrame("focus")
	end
	self:AddGroupToAceOptions("focustarget")
	if not self.db.profile.groups.focustarget.hidden then
		self:CreateUnitFrame("focustarget")
	end
	self:AddGroupToAceOptions("party")
	if not self.db.profile.groups.party.hidden then
		for i = 1, 4 do
			self:CreateUnitFrame('party' .. i)
		end
	end
	self:AddGroupToAceOptions("partypet")
	if not self.db.profile.groups.partypet.hidden then
		for i = 1, 4 do
			self:CreateUnitFrame('partypet' .. i)
		end
	end
	self:AddGroupToAceOptions("partytarget")
	if not self.db.profile.groups.partytarget.hidden then
		for i = 1, 4 do
			self:CreateUnitFrame('party' .. i .. 'target')
		end
	end
	self:AddGroupToAceOptions("raid")
	if not self.db.profile.groups.raid.hidden then
		for i = 1, 40 do
			self:CreateUnitFrame('raid' .. i)
		end
	end
	self:AddGroupToAceOptions("raidpet")
	if not self.db.profile.groups.raidpet.hidden then
		for i = 1, 40 do
			self:CreateUnitFrame('raidpet' .. i)
		end
	end
	self:AddGroupToAceOptions("raidtarget")
	if not self.db.profile.groups.raidtarget.hidden then
		for i = 1, 40 do
			self:CreateUnitFrame('raid' .. i .. 'target')
		end
	end
	self:PARTY_MEMBERS_CHANGED()
	self:PLAYER_PET_CHANGED()
	
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("UNIT_PET", "PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "UpdateAll")
	self:RegisterEvent("AceEvent_FullyInitialized", "UpdateAll")
	self:RegisterEvent("PLAYER_PET_CHANGED")
	self:RegisterEvent("SharedMedia_Registered")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	
	self:ScheduleRepeatingEvent("PitBull-UpdateWackyFrames", self.UpdateWackyFrames, 0.15, self)
	
	self:RegisterChatCommand('/PitBull', '/Pit', {
		type = 'execute',
		func = function()
			Dewdrop:Open("PitBull")
		end
	})
	
	Dewdrop:Register("PitBull",
		'children', PitBull.options,
		'cursorX', true,
		'cursorY', true
	)
end

local function getScale(group)
	return self.db.profile.groups[group].scale
end
local function getWidth(group)
	return self.db.profile.groups[group].width
end
local function getHeight(group)
	return self.db.profile.groups[group].height
end
local function getLayout(group)
	return self.db.profile.groups[group].layout
end
local function getBorder(group)
	return self.db.profile.groups[group].border
end
local function getHidden(group)
	return self.db.profile.groups[group].hidden
end
local function getDirection(group)
	return self.db.profile.groups[group].direction
end
local function getSpacing(group)
	return self.db.profile.groups[group].spacing
end
local function getFreeform(group)
	return self.db.profile.groups[group].freeform
end
function PitBull:AddGroupToAceOptions(group)
	if self.options.args[group] then
		return
	end
	groupMenus[group] = true
	local order
	if group:find("^player") then
		order = 50 + group:len()/1000
	elseif group:find("^pet") then
		order = 51 + group:len()/1000
	elseif group:find("^target") then
		order = 52 + group:len()/1000
	elseif group:find("^partypet") then
		order = 54 + group:len()/1000
	elseif group:find("^party") then
		order = 53 + group:len()/1000
	elseif group:find("^raidpet") then
		order = 56 + group:len()/1000
	elseif group:find("^raid") then
		order = 55 + group:len()/1000
	elseif group:find("^focus") then
		order = 57 + group:len()/1000
	elseif group:find("^mouseover") then
		order = 58 + group:len()/1000
	else
		order = 59 + group:len()/1000
	end
	self.options.args[group] = {
		order = order,
		name = GroupToLocale[group],
		desc = GroupToLocale[group],
		type = 'group',
		args = {
			header = {
				order = 1,
				name = GroupToLocale[group],
				type = 'header',
			},
			size = {
				name = "Size",
				desc = "Size of the unit frames",
				type = 'group',
				args = {
					scale = {
						name = "Scale",
						desc = "Scale of the unit frame",
						type = 'range',
						isPercent = true,
						min = 0.5,
						max = 2,
						step = 0.05,
						get = getScale,
						set = "ChangeScale",
						passValue = group,
						handler = PitBull,
						disabled = InCombatLockdown,
					},
					width = {
						name = "Width",
						desc = "Width of the unit frame",
						type = 'range',
						min = 20,
						max = 300,
						step = 5,
						get = getWidth,
						set = "ChangeWidth",
						passValue = group,
						handler = PitBull,
						disabled = InCombatLockdown,
					},
					height = {
						name = "Height",
						desc = "Height of the unit frame",
						type = 'range',
						min = 20,
						max = 300,
						step = 5,
						get = getHeight,
						set = "ChangeHeight",
						passValue = group,
						handler = PitBull,
						disabled = InCombatLockdown,
					},
				},
				disabled = InCombatLockdown,
			},
			layout = {
				name = "Layout",
				desc = "Layout",
				type = 'text',
				validate = layoutNames,
				get = getLayout,
				set = "ChangeLayout",
				passValue = group,
				handler = PitBull,
			},
			layoutSettings = {
				name = "Layout settings",
				desc = "Layout settings",
				type = 'group',
				args = {},
				hidden = function()
					if not next(self.options.args[group].args.layoutSettings.args) then
						return true
					end
					for k,v in pairs(self.options.args[group].args.layoutSettings.args) do
						if not v.hidden then
							return false
						elseif type(v.hidden) == "function" then
							local val = v.hidden(v.passValue)
							if not val then
								return false
							end
						elseif v.hidden ~= true then
							return false
						end
					end
					return true
				end
			},
			border = {
				name = "Border",
				desc = "Border",
				type = 'text',
				validate = borderNames,
				get = getBorder,
				set = "ChangeBorder",
				passValue = group,
				handler = PitBull,
			},
			hidden = {
				name = "Hidden",
				desc = "Hidden",
				type = 'toggle',
				get = getHidden,
				set = "ChangeHidden",
				passValue = group,
				handler = PitBull,
				disabled = InCombatLockdown
			}
		},
	}
	if group:find("^party") then
		self.options.args[group].args.grouping = {
			name = "Grouping",
			desc = "Grouping",
			type = 'group',
			args = {
				freeform = {
					name = "Enable freeform movement",
					desc = "Enable freeform movement",
					type = 'toggle',
					get = getFreeform,
					set = "ChangeGroupFreeform",
					passValue = group,
					handler = PitBull,
					disabled = InCombatLockdown,
				},
				direction = {
					name = "Direction",
					desc = "Direction",
					type = 'text',
					validate = {
						left = "Left",
						right = "Right",
						up = "Up",
						down = "Down",
					},
					get = getDirection,
					set = "ChangeGroupDirection",
					passValue = group,
					handler = PitBull,
					disabled = InCombatLockdown
				},
				spacing = {
					name = "Spacing",
					desc = "Spacing",
					type = 'range',
					min = 0,
					max = 200,
					step = 5,
					get = getSpacing,
					set = "ChangeGroupSpacing",
					passValue = group,
					handler = PitBull,
					disabled = InCombatLockdown
				},
			}
		}
	end
	if not self.options.args.sep then
		self.options.args.sep = {
			order = 60,
			type = 'header'
		}
	end
	
	for module, method in pairs(moduleOptionsMethods) do
		local name = module.name
	
		local options = PitBull.options.args[group].args
		if type(method) == "function" then
			options[name] = method(group)
		else
			options[name] = module[method](module, group)
		end
		if options[name] and not options[name].handler then
			options[name].handler = module
		end
	end
end

function PitBull:OnModuleCreated(name, module)
	for k in pairs(modulesWithFunction) do
		modulesWithFunction[k] = nil
	end
end

function PitBull:OnTooltipUpdate()
	Tablet:SetHint("|cffffff00Right-click|r to change settings")
end

local tmp = {}
function PitBull:ReorganizeCluster(cluster)
	local frames = allFramesByCluster[cluster]
	local _,f = next(frames)
	local group = f.group
	local freeform = self.db.profile.groups[group].freeform
	if freeform then
		for unit, frame in pairs(frames) do
			if unit == "head" then
				unit = next(frames, unit)
				if unit == nil then
					unit = next(frames, nil)
				end
				unit = unit:gsub("%d+", "1")
			end
			local uis = UIParent:GetScale()
			local s = frame:GetEffectiveScale()
			local db = self.db.profile.units[unit]
			frame:ClearAllPoints()
			frame:SetPoint("CENTER", db.x*uis/s, db.y*uis/s)
		end
	else
		for unit, frame in pairs(frames) do
			if unit ~= "head" then
				tmp[#tmp+1] = unit
			end
		end
		table.sort(tmp)
		local headFrame = frames.head
		local uis = UIParent:GetScale()
		local s = headFrame:GetEffectiveScale()
		local db = self.db.profile.units[tmp[1]]
		headFrame:ClearAllPoints()
		headFrame:SetPoint("CENTER", db.x*uis/s, db.y*uis/s)
	
		local direction = self.db.profile.groups[group].direction
		local spacing = self.db.profile.groups[group].spacing
		local size
		if direction == "left" or direction == "right" then
			size = headFrame:GetWidth()
		else
			size = headFrame:GetHeight()
		end
		for i = 1, #tmp do
			local unit = tmp[i]
			local frame = frames[unit]
			frame:ClearAllPoints()
			if direction == "left" then
				frame:SetPoint("RIGHT", headFrame, "RIGHT", (-size - spacing)*(i-1), 0)
			elseif direction == "right" then
				frame:SetPoint("LEFT", headFrame, "LEFT", (size + spacing)*(i-1), 0)
			elseif direction == "up" then
				frame:SetPoint("BOTTOM", headFrame, "BOTTOM", 0, (size + spacing)*(i-1))
			else
				frame:SetPoint("TOP", headFrame, "TOP", 0, (-size - spacing)*(i-1))
			end
			tmp[i] = nil
		end
	end
end

function PitBull:PLAYER_REGEN_DISABLED()
	if configMode then
		self:Print("Turning off config mode due to entering combat. Feel free to re-enable it once you are out of combat.")
		
		self:ChangeConfigMode(nil)
	end
end

function PitBull:PLAYER_TARGET_CHANGED()
	if configMode or UnitExists('target') then
		self:PopulateUnitFrame('target')
		if configMode or UnitExists('targettarget') then
			self:PopulateUnitFrame('targettarget')
			if configMode or UnitExists('targettargettarget') then
				self:PopulateUnitFrame('targettargettarget')
			end
		end
	end
end

function PitBull:PARTY_MEMBERS_CHANGED() -- also RAID_ROSTER_UPDATE and UNIT_PET
	if configMode then
		if configMode == "raid" then
			if GetNumRaidMembers() == 0 then
				self:Print("Turning off Configuration mode, due to real group change. Feel free to re-enable.")
				self:ChangeConfigMode(nil)
			end
		elseif configMode == "party" then
			if GetNumRaidMembers() > 0 or GetNumPartyMembers() == 0 then
				self:Print("Turning off Configuration mode, due to real group change. Feel free to re-enable.")
				self:ChangeConfigMode(nil)
			end
		elseif configMode == "solo" then
			if GetNumRaidMembers() > 0 or GetNumPartyMembers() > 0 then
				self:Print("Turning off Configuration mode, due to real group change. Feel free to re-enable.")
				self:ChangeConfigMode(nil)
			end
		end
	end
	local raid = GetNumRaidMembers()
	local hidePartyInRaid = self.db.profile.settings.hidePartyInRaid
	if raid > 0 then
		for unit, frame in pairs(allFrames) do
			if unit:find("^party") then
				if hidePartyInRaid then
					if frames[unit] then
						self:ClearUnitFrame(unit)
					end
					self:ScheduleLeaveCombatAction(frame, 'EnableMouse', false)
				else
					if configMode or UnitExists(unit) then
						self:PopulateUnitFrame(unit)
					else
						self:ClearUnitFrame(unit)
					end
				end
			elseif unit:find("^raid") then
				if configMode or UnitExists(unit) then
					self:PopulateUnitFrame(unit)
				else
					self:ClearUnitFrame(unit)
				end
			end
		end
	else
		for unit, frame in pairs(allFrames) do
			if unit:find("^party") then
				self:ScheduleLeaveCombatAction(frame, 'EnableMouse', true)
				if configMode or UnitExists(unit) then
					self:PopulateUnitFrame(unit)
				else
					self:ClearUnitFrame(unit)
				end
			elseif unit:find("^raid") then
				self:ClearUnitFrame(unit)
			end
		end
	end
	
	if configMode or UnitExists('pet') then
		self:PopulateUnitFrame('pet')
	else
		self:ClearUnitFrame('pet')
	end
end

function PitBull:PLAYER_PET_CHANGED()
	if configMode or UnitExists('pet') then
		self:PopulateUnitFrame('pet')
	else
		self:ClearUnitFrame('pet')
	end
end

function PitBull:UpdateAll()
	for unit in pairs(frames) do
		if configMode or UnitExists(unit) then
			self:UpdateLayout(unit)
			self:UpdateBackdropAndBorder(unit)
			self:UpdateFrame(unit)
		end
	end
	self:UpdateAllStatusBarTextures()
end

function PitBull:UpdateWackyFrames()
	-- this is to update frames that are not handled by Blizzard's standard events, e.g. targettarget, targettargettarget
	for i = 1, wackyUnits.n do
		local unit = wackyUnits[i]
		if configMode or UnitExists(unit) then
			self:UpdateFrame(unit)
		end
	end
end

function PitBull:UpdateFrame(unit)
	-- update everything about a frame
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `UpdateFrame'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if not frame then
		return
	end
	
	for _, module in ipairs(modulesWithFunction.OnUpdateFrame) do
		module:OnUpdateFrame(unit, frame)
	end
	
	for text, u in pairs(self.watchdogRegistry) do
		if unit == u then
			self:UpdateWatchdogFontString(text)
		end
	end
end

local function isframe(frame)
	return type(frame) == "table" and type(rawget(frame, 0)) == "userdata" and type(frame.IsObjectType) == "function"
end

local tmp = {}
local ignore = { overlay = true, db = true, group = true, unit = true }
function PitBull:UpdateLayout(unit)
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `UpdateLayout'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if not frame then
		return
	end
	
	local layoutName = self.db.profile.groups[frame.group].layout
	local current_layout = layouts[layoutName]
	if not current_layout then
		layoutName = "ABF"
		current_layout = layouts.ABF
	end
	
	if current_layout.extraFrames then
		for _, v in ipairs(current_layout.extraFrames) do
			if frame[v] then
				frame[v] = delFrame(frame[v])
			end
		end
	end
	
	for k, v in pairs(frame) do
		if not ignore[k] and type(k) == "string" and isframe(v) then
			tmp[k] = v:IsShown()
			v:Show()
			v:ClearAllPoints()
			v:SetWidth(0)
			v:SetHeight(0)
			if v.SetScale then
				v:SetScale(1)
			end
			local objectType = v:GetObjectType()
			if objectType == "FontString" then
				local fontObject = v:GetFontObject()
				if fontObject then
					v:SetFont(fontObject:GetFont())
				end
				v:SetJustifyH("LEFT")
				v:SetHeight(1)
			elseif objectType == "StatusBar" then
				v:SetOrientation("HORIZONTAL")
			end
		end
	end
	for k, v in pairs(frame) do
		if not ignore[k] and type(k) == "string" and isframe(v) then
			local func = current_layout[k]
			if type(func) == "function" then
				local success, ret = pcall(func, unit, frame)
				if not success then
					self:Print("There was an error with the layout %q, preventing the frame for %q to be shown properly. Please inform the author: %s", layoutName, unit, ret)
					v:Hide()
				end
			else
				local module = frameToModule[k]
				if module and type(module.OnUnknownLayout) == "function" then
					module:OnUnknownLayout(unit, frame, k)
				end
			end
			if tmp[k] ~= v:IsShown() then
				-- change of visibility
				if v:IsObjectType("FontString") then
					if v:IsShown() then
						self:UpdateWatchdogFontString(v)
					end
				end
			end
			tmp[k] = nil
		end
	end
	for _, module in ipairs(modulesWithFunction.OnUpdateLayout) do
		module:OnUpdateLayout(unit, frame)
	end
	
	self:UpdateLayoutSettings(frame.group)
end

function PitBull:UpdateAllLayouts()
	for unit, frame in pairs(frames) do
		self:UpdateLayout(unit)
	end
end

function PitBull:UpdateBackdropAndBorder(unit)
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `UpdateBackdropAndBorder'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if not frame then
		return
	end
	local border = borders[self.db.profile.groups[frame.group].border] or borders.Hidden
	local edgeFile = border[1] or ""
	local edgeSize = border[2] or 0
	local inset = border[3] or 3
	local bg = newHash(
		'tile', true,
		'tileSize', 16,
		'bgFile', "Interface\\ChatFrame\\ChatFrameBackground",
		'edgeFile', edgeFile,
		'edgeSize', edgeSize,
		'insets', newHash(
			'left', inset,
			'right', inset,
			'top', inset,
			'bottom', inset
		)
	)
	frame:SetBackdrop(bg)
	bg.insets = del(bg.insets)
	bg = del(bg)
	frame:SetBackdropColor(unpack(self.db.profile.settings.colors.frameBG))
	frame:SetBackdropBorderColor(unpack(self.db.profile.settings.colors.frameBorder))
end

function PitBull:UpdateAllBackdropsAndBorders()
	for unit, frame in pairs(frames) do
		self:UpdateBackdropAndBorder(unit)
	end
end

function PitBull:AreUnitFramesClampedToScreen()
	return self.db.profile.settings.clamped
end

function PitBull:ToggleUnitFramesClampedToScreen(value)
	if value == nil then
		value = not self.db.profile.settings.clamped
	end
	self.db.profile.settings.clamped = value
	for unit, frame in pairs(allFrames) do
		frame:SetClampedToScreen(value)
	end
end

function PitBull:PopulateUnitFrame(unit)
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `PopulateUnitFrame'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if frame then
		self:UpdateFrame(unit)
		return
	else
		frame = allFrames[unit]
		if not frame then
			return
		end
	end
	
	local group = frame.group
	self:AddGroupToAceOptions(group)
	
	local overlay = newFrame("Frame", frame)
	frame.overlay = overlay
	overlay:SetFrameLevel(frame:GetFrameLevel()+3)
	
	for _, module in ipairs(modulesWithFunction.OnPopulateUnitFrame) do
		module:OnPopulateUnitFrame(unit, frame)
	end
	
	frames[unit] = frame
	if not framesByGroup[group] then
		framesByGroup[group] = new()
	end
	framesByGroup[group][unit] = frame
	if not IsWackyUnit[unit] then
		nonWackyFrames[unit] = frame
	end
	
	self:UpdateLayout(unit)
	self:UpdateBackdropAndBorder(unit)
	self:UpdateFrame(unit)
end

function PitBull:ClearUnitFrame(unit)
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `ClearUnitFrame'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if not frame then
		return
	end
	
	local group = frame.group
	
	local current_layout = layouts[self.db.profile.groups[frame.group].layout] or layouts.ABF
	
	if current_layout.extraFrames then
		for _, v in ipairs(current_layout.extraFrames) do
			if frame[v] then
				frame[v] = delFrame(frame[v])
			end
		end
	end
	
	for _, module in ipairs(modulesWithFunction.OnClearUnitFrame) do
		module:OnClearUnitFrame(unit, frame)
	end
	
	frame.overlay = delFrame(frame.overlay)
	
	local bg = new()
	frame:SetBackdrop(bg)
	bg = del(bg)
	frame:SetBackdropColor(0, 0, 0, 0)
	frame:SetBackdropBorderColor(0, 0, 0, 0)
	
	frames[unit] = nil
	framesByGroup[group][unit] = nil
	if not next(framesByGroup[group]) then
		framesByGroup[group] = del(framesByGroup[group])
	end
	nonWackyFrames[unit] = nil
end

local function frame_custom_menu(this, unit)
	Dewdrop:Open(this,
		'children', PitBull.options.args[this.group],
		'cursorX', true,
		'cursorY', true
	)
end
frame_custom_menu = wrap(frame_custom_menu, '/frame_custom_menu')

local function frame_menu(this, unit)
	local type, num
	
	if unit == "player" then
		type = PlayerFrameDropDown
	elseif unit == "target" then
		type = TargetFrameDropDown
	elseif unit == "pet" then
		type = PetFrameDropDown
	else
		num = unit:match("^party(%d)$")
		if num then
			type = _G["PartyMemberFrame" .. num .. "DropDown"]
		else
			num = unit:match("^raid(%d%d?)$")
			if num then
				type = FriendsDropDown
				this.unit = unit
				this.name = UnitName(unit)
				this.id = this:GetID()
				FriendsDropDown.displayMode = "MENU"
				FriendsDropDown.initialize = RaidFrameDropDown_Initialize
			else
				return frame_custom_menu(this, unit)
			end
		end
	end
	
	if num then
		PitBull:ScheduleLeaveCombatAction(this, "SetID", tonumber(num))
	end

	if type then
		HideDropDownMenu(1)
		type.unit = unit
		type.name = UnitName(unit)
		ToggleDropDownMenu(1, nil, type, "cursor")
		return true
	end
end
frame_menu = wrap(frame_menu, '/frame_menu')

local function frame_OnEnter(this)
	if not frames[this.unit] then
		return
	end
	for _, module in ipairs(modulesWithFunction.OnFrameOnEnter) do
		module:OnFrameOnEnter(this.unit, this)
	end
	UnitFrame_OnEnter()
end
frame_OnEnter = wrap(frame_OnEnter, '/frame_OnEnter')

local function frame_OnLeave(this)
	if not frames[this.unit] then
		return
	end
	for _, module in ipairs(modulesWithFunction.OnFrameOnLeave) do
		module:OnFrameOnLeave(this.unit, this)
	end
	UnitFrame_OnLeave()
end
frame_OnLeave = wrap(frame_OnLeave, '/frame_OnLeave')

local function frame_OnDragStart(this)
	if PitBull.db.profile.settings.locked and not IsAltKeyDown() then
		return
	end
	if not frames[this.unit] then
		return
	end
	-- also want to move group if grouped
	local cluster = this.cluster
	if cluster then
		allFramesByCluster[cluster].head:StartMoving()
	else
		this:StartMoving()
	end
end
frame_OnDragStart = wrap(frame_OnDragStart, '/frame_OnDragStart')

local function frame_OnDragStop(this)
	if not frames[this.unit] then
		return
	end
	
	local cluster = this.cluster
	if cluster then
		this = allFramesByCluster[cluster].head
	end
	
	this:StopMovingOrSizing()

	local unit = this.unit
	local db = self.db.profile.units[unit]
	local x, y = this:GetCenter()
	local s = this:GetEffectiveScale()
	local uis = UIParent:GetScale()
	x = x*s - GetScreenWidth()*uis/2
	y = y*s - GetScreenHeight()*uis/2
	db.x, db.y = x/uis, y/uis
	this:ClearAllPoints()
	this:SetPoint("CENTER", x/s, y/s)
	
	if cluster then
		PitBull:ReorganizeCluster(cluster)
	end
end
frame_OnDragStop = wrap(frame_OnDragStop, '/frame_OnDragStop')

local function frame_OnShow(this)
	local unit = this.unit
	local raid = GetNumRaidMembers()
	if raid > 0 and unit:find("^party") then
		return
	end
	PitBull:PopulateUnitFrame(unit)
end
frame_OnShow = wrap(frame_OnShow, '/frame_OnShow')

local function frame_OnHide(this)
	PitBull:ClearUnitFrame(this.unit)
end
frame_OnHide = wrap(frame_OnHide, '/frame_OnHide')

local count = 0
local cache = {}
function PitBull:CreateUnitFrame(unit)
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `CreateUnitFrame'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if frame then
		self:UpdateFrame(unit)
		return frame
	end
	
	if not IsLegitimateUnit[unit] then
		error(("Bad argument #2 to `CreateUnitFrame': %q is not a legitimate unit"):format(unit), 2)
	end
	
	count = count + 1
	if #cache == 0 then
		local name = "PitBullUnitFrame" .. count
		frame = CreateFrame("Button", name, UIParent, "SecureUnitButtonTemplate")
		frame:Hide()
		frame:EnableMouse(true)
		frame:SetScript("OnEnter", frame_OnEnter)
		frame:SetScript("OnLeave", frame_OnLeave)
		frame:SetScript("OnDragStart", frame_OnDragStart)
		frame:SetScript("OnDragStop", frame_OnDragStop)
		frame:SetScript("OnHide", frame_OnHide)
		frame:SetScript("OnShow", frame_OnShow)
		frame:SetMovable(true)
		frame:RegisterForDrag("LeftButton")
		frame:RegisterForClicks("LeftButtonUp","RightButtonUp","MiddleButtonUp","Button4Up","Button5Up")
		frame:SetAttribute("type1", "target")
		frame:SetAttribute("type2", "menu")
		frame:SetAttribute("*type2", "custom_menu")
		frame.menu = frame_menu
		frame.custom_menu = frame_custom_menu
		frame:SetClampedToScreen(self.db.profile.settings.clamped)
	else
		frame = table_remove(cache)
	end
	
	local group = unit:gsub("%d", "")
	frame.group = group
	local cluster
	if unit:find("^party") then
		cluster = unit:gsub("%d", "")
	end
	
	allFrames[unit] = frame
	if not allFramesByGroup[group] then
		allFramesByGroup[group] = new()
	end
	allFramesByGroup[group][unit] = frame
	if cluster then
		frame.cluster = cluster
		if not allFramesByCluster[cluster] then
			allFramesByCluster[cluster] = new()
			local head = newFrame("Frame", UIParent)
			allFramesByCluster[cluster].head = head
			partyhead = head
			head.unit = unit:gsub("%d+", "1")
			head.group = group
			local db = self.db.profile.units[head.unit]
			head:SetWidth(self.db.profile.groups[group].width)
			head:SetHeight(self.db.profile.groups[group].height)
			head:SetScale(self.db.profile.groups[group].scale)
			local uis = UIParent:GetScale()
			local s = frame:GetEffectiveScale()
			local num = tonumber(unit:match("(%d+)"))
			head:SetPoint("CENTER", db.x*uis/s, db.y*uis/s)
			head:SetMovable(true)
		end
		allFramesByCluster[cluster][unit] = frame
	end
	
	-- TODO: get rid of this section eventually, replace with local db = self.db.profile.units[unit] --[[
	local db = self.db.profile[unit]
	if db then
		self.db.profile[unit] = nil
		local x, y = db.x, db.y
		db = self.db.profile.units[unit]
		db.x, db.y = x, y
	else
		db = self.db.profile.units[unit]
	end
	--]] TODO: get rid of this section eventually, replace with local db = self.db.profile.units[unit] --
	
	frame:SetWidth(self.db.profile.groups[group].width)
	frame:SetHeight(self.db.profile.groups[group].height)
	frame:SetScale(self.db.profile.groups[group].scale)

	-- db.{x,y} are scaled coordinates, in the UIParent scale.  Switch from
	-- that scale to the frame's scale.
	if cluster then
		self:ReorganizeCluster(cluster)
	else
		local uis = UIParent:GetScale()
		local s = frame:GetEffectiveScale()
		local num = tonumber(unit:match("(%d+)"))
		frame:SetPoint("CENTER", db.x*uis/s, db.y*uis/s)
	end
	
	frame.unit = unit
	frame:SetAttribute("unit", unit)
	
	if IsWackyUnit[unit] then
		wackyUnits.n = wackyUnits.n + 1
		wackyUnits[wackyUnits.n] = unit
	end
	
	if configMode or UnitExists(unit) then
		self:PopulateUnitFrame(unit)
	end
	
	ClickCastFrames[frame] = true
	
	if not configMode then
		RegisterUnitWatch(frame)
	else
		frame:Show()
	end
	
	return frame
end

function PitBull:DestroyUnitFrame(unit)
	if type(unit) ~= "string" then
		error(("Bad argument #2 to `DestroyUnitFrame'. Expected %s, got %s"):format("string", type(unit)), 2)
	end
	local frame = frames[unit]
	if not frame then
		return
	end
	self:ClearUnitFrame(unit)
	cache[#cache+1] = frame
	UnregisterUnitWatch(frame)
	frame:Hide()
	frame.unit = 'none'
	local group = frame.group
	frame.group = nil
	frame:SetAttribute("unit", 'none')
	frames[unit] = nil
	allFrames[unit] = nil
	if framesByGroup[group] then
		framesByGroup[group][unit] = nil
		if not next(framesByGroup[group]) then
			framesByGroup[group] = del(framesByGroup[group])
		end
	end
	allFramesByGroup[group][unit] = nil
	if not next(allFramesByGroup[group]) then
		allFramesByGroup[group] = del(allFramesByGroup[group])
	end
	if frame.cluster then
		local cluster = frame.cluster
		frame.cluster = nil
		framesByCluster[cluster][unit] = nil
		if not next(framesByCluster[cluster]) then
			framesByCluster[cluster] = del(framesByCluster[cluster])
		end
	end
	nonWackyFrames[unit] = nil
	if IsWackyUnit[unit] then
		for i,v in ipairs(wackyUnits) do
			if v == unit then
				table_remove(wackyUnits, i)
				break
			end
		end
		wackyUnits.n = wackyUnits.n - 1
	end
end

function PitBull:ChangeScale(group, value)
	if self.db.profile.groups[group].scale == value then
		return
	end
	self.db.profile.groups[group].scale = value
	
	if allFramesByGroup[group] then
		for unit, frame in pairs(allFramesByGroup[group]) do
			frame:SetScale(value)
			local cluster = frame.cluster
			if cluster then
				allFramesByCluster[cluster].head:SetScale(value)
			end
		
			if frames[unit] then
				self:UpdateLayout(unit)
			end
		end
	end
end

function PitBull:ChangeWidth(group, value)
	if self.db.profile.groups[group].width == value then
		return
	end
	self.db.profile.groups[group].width = value
	
	if allFramesByGroup[group] then
		for unit, frame in pairs(allFramesByGroup[group]) do
			frame:SetWidth(value)
			local cluster = frame.cluster
			if cluster then
				allFramesByCluster[cluster].head:SetWidth(value)
			end
		
			if frames[unit] then
				self:UpdateLayout(unit)
			end
		end
	end
end

function PitBull:ChangeHeight(group, value)
	if self.db.profile.groups[group].height == value then
		return
	end
	self.db.profile.groups[group].height = value
	
	if allFramesByGroup[group] then
		for unit, frame in pairs(allFramesByGroup[group]) do
			frame:SetHeight(value)
			local cluster = frame.cluster
			if cluster then
				allFramesByCluster[cluster].head:SetHeight(value)
			end
		
			if frames[unit] then
				self:UpdateLayout(unit)
			end
		end
	end
end

function PitBull:GetStatusBarTexture()
	return SharedMedia:Fetch('statusbar', self.db.profile.settings.texture)
end

function PitBull:UpdateAllStatusBarTextures()
	local texture = PitBull:GetStatusBarTexture()
	for _,module in ipairs(modulesWithFunction.OnUpdateStatusBarTexture) do
		module:OnUpdateStatusBarTexture(texture)
	end
end

function PitBull:SharedMedia_Registered(kind, name)
	if kind == "statusbar" then
		if self.db.profile and name == self.db.profile.settings.texture then
			self:UpdateAllStatusBarTextures()
		end
	end
end

function PitBull:GetUnitFrame(unit)
	return frames[unit]
end

function PitBull:IterateUnitFrames()
	return pairs(frames)
end

local function blankIter() end
function PitBull:IterateUnitFramesByGroup(group)
	local g = framesByGroup[group]
	if not g then
		return blankIter
	else
		return pairs(g)
	end
end

function PitBull:IterateNonWackyUnitFrames()
	return pairs(nonWackyFrames)
end

function PitBull:OnProfileEnable()
	self.colorConstants = self.db.profile.settings.colors
	self:UpdateAllStatusBarTextures()
	for _,module in ipairs(modulesWithFunction.OnPitBullProfileEnable) do
		module:OnPitBullProfileEnable()
	end
	
	self:UpdateAll()
end

function PitBull:RegisterLayout(name, data)
	layouts[name] = data
	layoutNames[name] = name
	
	for unit, frame in self:IterateUnitFrames() do
		if self.db.profile.groups[frame.group].layout == name then
			self:UpdateLayout(unit)
		end
	end
end

function PitBull:RegisterBorder(name, texture, size, insets)
	borders[name] = { texture, size, insets }
	borderNames[name] = name
	
	for unit, frame in self:IterateUnitFrames() do
		if self.db.profile.groups[frame.group].border == name then
			self:UpdateBackdropAndBorder(unit)
		end
	end
end

function PitBull:UpdateLayoutSettings(group)
	local layout = layouts[self.db.profile.groups[group].layout] or layouts.ABF
	
	for k, v in pairs(layout) do
		if type(k) == "string" and type(v) == "function" then
			local name = k:match("^(.+)_options$")
			if name and not self.options.args[group].args.layoutSettings.args[name] then
				local options = v(group)
				self.options.args[group].args.layoutSettings.args[name] = options
				if not options.handler then
					options.handler = layout
				end
				local function func()
					local good = false
					for unit, frame in self:IterateUnitFramesByGroup(group) do
						if frame[name] then
							good = true
							break
						end
					end
					return not good
				end
				if options.hidden then
					if type(options.hidden) == "function" then
						local options_hidden = options.hidden
						function options.hidden(...)
							if options_hidden(...) then
								return true
							end
							return func()
						end
					elseif type(options.hidden) == "string" then
						local options_hidden = options.hidden
						function options.hidden(...)
							local handler = options.handler
							if handler[options_hidden](handler, ...) then
								return true
							end
							return func()
						end
					end
				else
					options.hidden = func
				end
			end
		end
	end
end

function PitBull:ChangeLayout(group, value)
	local old_layout = layouts[self.db.profile.groups[group].layout] or layouts.ABF
	local new_layout = layouts[value] or layouts.ABF
	
	self.db.profile.groups[group].layout = value
	if old_layout == new_layout then
		return
	end
	
	if old_layout.extraFrames then
		for unit, frame in self:IterateUnitFramesByGroup(group) do
			for _, v in ipairs(old_layout.extraFrames) do
				if frame[v] then
					frame[v] = delFrame(frame[v])
				end
			end
		end
	end
	for k, v in pairs(old_layout) do
		if type(k) == "string" and k:find("_options$") then
			self.options.args[group].args.layoutSettings.args[k] = nil
		end
	end
	
	for unit, frame in self:IterateUnitFramesByGroup(group) do
		self:UpdateLayout(unit)
	end
end

function PitBull:ChangeBorder(group, value)
	self.db.profile.groups[group].border = value
	
	for unit, frame in self:IterateUnitFramesByGroup(group) do
		self:UpdateBackdropAndBorder(unit)
	end
end

local tmp = {}
function PitBull:ChangeHidden(group, value)
	self.db.profile.groups[group].hidden = value
	
	if value then
		for unit, frame in pairs(allFramesByGroup[group]) do
			self:DestroyUnitFrame(unit)
		end
	else
		if group:find("^raidpet") then
			for i = 1, 40 do
				tmp['raidpet' .. i .. group:sub(8)] = true
			end
		elseif group:find("^raid") then
			for i = 1, 40 do
				tmp['raid' .. i .. group:sub(5)] = true
			end
		elseif group:find("^partypet") then
			for i = 1, 4 do
				tmp['partypet' .. i .. group:sub(9)] = true
			end
		elseif group:find("^party") then
			for i = 1, 4 do
				tmp['party' .. i .. group:sub(6)] = true
			end
		else
			tmp[group] = true
		end
		
		for unit in pairs(tmp) do
			self:CreateUnitFrame(unit)
			tmp[unit] = nil
		end
	end
end

function PitBull:ChangeGroupFreeform(group, value)
	self.db.profile.groups[group].freeform = value

	self:ReorganizeCluster(group)
end

function PitBull:ChangeGroupDirection(group, value)
	self.db.profile.groups[group].direction = value
	
	self:ReorganizeCluster(group)
end

function PitBull:ChangeGroupSpacing(group, value)
	self.db.profile.groups[group].spacing = value
	
	self:ReorganizeCluster(group)
end

function PitBull:ChangeHidePartyInRaid(value)
	self.db.profile.settings.hidePartyInRaid = value
	
	local raid = GetNumRaidMembers()
	if raid > 0 then
		for unit, frame in pairs(allFrames) do
			if unit:find("^party") then
				if value then
					if frames[unit] then
						self:ClearUnitFrame(unit)
					end
					self:ScheduleLeaveCombatAction(frame, 'EnableMouse', false)
				else
					if configMode or UnitExists(unit) then
						self:PopulateUnitFrame(unit)
					else
						self:ClearUnitFrame(unit)
					end
					self:ScheduleLeaveCombatAction(frame, 'EnableMouse', true)
				end
			end
		end
	end
	
	self:PARTY_MEMBERS_CHANGED()
	
	self:ChangeConfigMode(configMode)
end

local function frame_OnMouseDown(this, button)
	if not UnitExists(this.unit) and button == "RightButton" then
		this:custom_menu(this.unit)
	end
end
function PitBull:ChangeConfigMode(value)
	if value == "disabled" then
		value = nil
	else
		local raid = GetNumRaidMembers()
		if raid > 0 then
			value = "raid"
		else
			local party = GetNumPartyMembers()
			if party > 0 then
				value = "party"
			end
		end
	end
	self.configMode = value
	configMode = value
	
	for unit, frame in pairs(allFrames) do
		if configMode then
			if (value == "solo" and (unit:find("^party") or unit:find("^raid"))) or (value == "party" and unit:find("^raid")) or (value == "raid" and self.db.profile.settings.hidePartyInRaid and unit:find("^party")) then
				RegisterUnitWatch(frame)
				frame:SetScript("OnMouseDown", nil)
			else
				UnregisterUnitWatch(frame)
				frame:Show()
				frame:SetScript("OnMouseDown", frame_OnMouseDown)
			end
		else
			RegisterUnitWatch(frame)
			frame:SetScript("OnMouseDown", nil)
		end
	end
	
	for _, module in ipairs(modulesWithFunction.OnChangeConfigMode) do
		module:OnChangeConfigMode(value)
	end
end

function PitBull.modulePrototype:RegisterPitBullChildFrames(...)
	for i = 1, select('#', ...) do
		local v = select(i, ...)
		if type(v) ~= "string" then
			error(("Bad argument #%d to `RegisterPitBullChildFrames'. Expected \"string\", got %q"):format(i, type(v)), 2)
		end
		if frameToModule[v] and frameToModule[v] ~= self then
			error(("Frame %q provided twice to `RegisterPitBullChildFrames'. Previously by %q and now by %q"):format(v, tostring(frameToModule[v]), tostring(self)), 2)
		end
		frameToModule[v] = self
	end
end

function PitBull.modulePrototype:RegisterPitBullOptionsMethod(method)
	local kind = type(method)
	if kind ~= "string" and kind ~= "function" then
		error(("Bad argument #2 to `RegisterPitBullOptionsMethod'. %s expected, got %s."):format("string or function", kind), 2)
	end
	if kind == "string" then
		if type(self[method]) ~= "function" then
			error(("Cannot find handler %q for `RegisterPitBullOptionsMethod'"):format(method), 2)
		end
	end
	
	if moduleOptionsMethods[self] then
		error("Cannot call `RegisterPitBullOptionsMethod' more than once from the same module.", 2)
	end
	
	local name = self.name
	
	for group in pairs(groupMenus) do
		local options = PitBull.options.args[group].args
		if kind == "function" then
			options[name] = method(group)
		else
			options[name] = self[method](self, group)
		end
		if options[name] and not options[name].handler then
			options[name].handler = self
		end
	end
	
	moduleOptionsMethods[self] = method
end

wrap(PitBull, '')

PitBull:RegisterBorder("Classic", "Interface\\Tooltips\\UI-Tooltip-Border", 16, 5)
PitBull:RegisterBorder("Nurfed", "Interface\\DialogFrame\\UI-DialogBox-Border", 16, 5)
PitBull:RegisterBorder("Hidden", "", 0, 3)

if DEBUG then
	local lastOpen = false
	function ProfilePitBull()
		if not Tablet:IsRegistered("PitBull") then
			local currentSort = 'time'
			local duration, times, memories, calls
			local sorts = {
				time = function(alpha, bravo)
					if times[alpha] ~= times[bravo] then
						return times[alpha] < times[bravo]
					elseif memories[alpha] ~= memories[bravo] then	
						return memories[alpha] < memories[bravo]
					elseif calls[alpha] ~= calls[bravo] then
						return calls[alpha] < calls[bravo]
					else
						return alpha:lower() < bravo:lower()
					end
				end,
				memory = function(alpha, bravo)
					if memories[alpha] ~= memories[bravo] then	
						return memories[alpha] < memories[bravo]
					elseif times[alpha] ~= times[bravo] then
						return times[alpha] < times[bravo]
					elseif calls[alpha] ~= calls[bravo] then
						return calls[alpha] < calls[bravo]
					else
						return alpha:lower() < bravo:lower()
					end
				end,
				calls = function(alpha, bravo)
					if calls[alpha] ~= calls[bravo] then
						return calls[alpha] < calls[bravo]
					elseif times[alpha] ~= times[bravo] then
						return times[alpha] < times[bravo]
					elseif memories[alpha] ~= memories[bravo] then	
						return memories[alpha] < memories[bravo]
					else
						return alpha:lower() < bravo:lower()
					end
				end,
				name = function(alpha, bravo)
					return alpha:lower() < bravo:lower()
				end,
			}
			Tablet:Register("PitBull",
				'cantAttach', true,
				'menu', {
					type = 'group',
					args = {
						time = {
							name = "Sort by time",
							desc = "Sort by time",
							type = 'toggle',
							get = function()
								return currentSort == 'time'
							end,
							set = function(value)
								currentSort = 'time'
							end,
						},
						memory = {
							name = "Sort by memory",
							desc = "Sort by memory",
							type = 'toggle',
							get = function()
								return currentSort == 'memory'
							end,
							set = function(value)
								currentSort = 'memory'
							end,
						},
						name = {
							name = "Sort by name",
							desc = "Sort by name",
							type = 'toggle',
							get = function()
								return currentSort == 'name'
							end,
							set = function(value)
								currentSort = 'name'
							end,
						},
						calls = {
							name = "Sort by calls",
							desc = "Sort by calls",
							type = 'toggle',
							get = function()
								return currentSort == 'calls'
							end,
							set = function(value)
								currentSort = 'calls'
							end,
						},
					},
				},
				'children', function()
					Tablet:SetTitle("PitBull Profile")
					
					local cat = Tablet:AddCategory(
						'columns', 7,
						'text', "Function",
						'text2', "Time",
						'text3', "Percent Time",
						'text4', "Memory",
						'text5', "Calls",
						'text6', "Time/Call",
						'text7', "Memory/Call"
					)
					duration, times, memories, calls = DEBUG()
					if not duration or not time or not memories or not calls then
						cat:AddLine('text', "Problems in debug")
						return
					end
					
					
					local t = new()
					for method in pairs(memories) do
						t[#t+1] = method
					end
					table.sort(t, sorts[currentSort])
					local memory = 0
					local time = 0
					local calls_total = 0
					for _,method in ipairs(t) do
						cat:AddLine(
							'text', method,
							'text2', ("%.3f s"):format(times[method]),
							'text3', ("%.3f%%"):format(times[method] / duration * 100),
							'text4', ("%.3f KiB"):format(memories[method]),
							'text5', ("%dx"):format(calls[method]),
							'text6', ("%.3f s"):format(times[method] / calls[method]),
							'text7', ("%.3f KiB"):format(memories[method] / calls[method])
						)
						memory = memory + memories[method]
						time = time + times[method]
						calls_total = calls_total + calls[method]
					end
					cat:AddLine(
						'text', ' '
					)
					cat:AddLine(
						'text', "Total",
						'text2', ("%.3f s"):format(time),
						'text3', ("%.3f%%"):format(time / duration * 100),
						'text4', ("%.3f KiB"):format(memory),
						'text5', ("%dx"):format(calls_total),
						'text6', ("%.3f s"):format(time / calls_total),
						'text7', ("%.3f KiB"):format(memory / calls_total)
					)
					del(t)
				end
			)
			lastOpen = true
			PitBull:ScheduleRepeatingEvent("PitBull-UpdateProfileTablet", function()
				Tablet:Refresh("PitBull")
			end, 0.2)
			Tablet:Open("PitBull")
		else
			if lastOpen then
				Tablet:Close("PitBull")
				lastOpen = false
			else
				Tablet:Open("PitBull")
				lastOpen = true
			end
		end
	end
end
